"""
This module initializes the `python_utils` package by importing various
submodules and functions.

Submodules:
    aio
    converters
    decorators
    formatters
    generators
    import_
    logger
    terminal
    time
    types

Functions:
    acount
    remap
    scale_1024
    to_float
    to_int
    to_str
    to_unicode
    listify
    set_attributes
    raise_exception
    reraise
    camel_to_underscore
    timesince
    abatcher
    batcher
    import_global
    get_terminal_size
    aio_generator_timeout_detector
    aio_generator_timeout_detector_decorator
    aio_timeout_generator
    delta_to_seconds
    delta_to_seconds_or_none
    format_time
    timedelta_to_seconds
    timeout_generator

Classes:
    CastedDict
    LazyCastedDict
    UniqueList
    Logged
    LoggerBase
"""

from . import (
    aio,
    converters,
    decorators,
    formatters,
    generators,
    import_,
    logger,
    terminal,
    time,
    types,
)
from .aio import acount
from .containers import CastedDict, LazyCastedDict, UniqueList
from .converters import remap, scale_1024, to_float, to_int, to_str, to_unicode
from .decorators import listify, set_attributes
from .exceptions import raise_exception, reraise
from .formatters import camel_to_underscore, timesince
from .generators import abatcher, batcher
from .import_ import import_global
from .logger import Logged, LoggerBase
from .terminal import get_terminal_size
from .time import (
    aio_generator_timeout_detector,
    aio_generator_timeout_detector_decorator,
    aio_timeout_generator,
    delta_to_seconds,
    delta_to_seconds_or_none,
    format_time,
    timedelta_to_seconds,
    timeout_generator,
)

__all__ = [
    'CastedDict',
    'LazyCastedDict',
    'Logged',
    'LoggerBase',
    'UniqueList',
    'abatcher',
    'acount',
    'aio',
    'aio_generator_timeout_detector',
    'aio_generator_timeout_detector_decorator',
    'aio_timeout_generator',
    'batcher',
    'camel_to_underscore',
    'converters',
    'decorators',
    'delta_to_seconds',
    'delta_to_seconds_or_none',
    'format_time',
    'formatters',
    'generators',
    'get_terminal_size',
    'import_',
    'import_global',
    'listify',
    'logger',
    'raise_exception',
    'remap',
    'reraise',
    'scale_1024',
    'set_attributes',
    'terminal',
    'time',
    'timedelta_to_seconds',
    'timeout_generator',
    'timesince',
    'to_float',
    'to_int',
    'to_str',
    'to_unicode',
    'types',
]



"""
This module provides custom container classes with enhanced functionality.

Classes:
    CastedDictBase: Abstract base class for dictionaries that cast keys and
        values.
    CastedDict: Dictionary that casts keys and values to specified types.
    LazyCastedDict: Dictionary that lazily casts values to specified types upon
        access.
    UniqueList: List that only allows unique values, with configurable behavior
        on duplicates.
    SliceableDeque: Deque that supports slicing and enhanced equality checks.

Type Aliases:
    KT: Type variable for dictionary keys.
    VT: Type variable for dictionary values.
    DT: Type alias for a dictionary with keys of type KT and values of type VT.
    KT_cast: Type alias for a callable that casts dictionary keys.
    VT_cast: Type alias for a callable that casts dictionary values.
    HT: Type variable for hashable values in UniqueList.
    T: Type variable for generic types.
    DictUpdateArgs: Union type for arguments that can be used to update a
        dictionary.
    OnDuplicate: Literal type for handling duplicate values in UniqueList.

Usage:
    - CastedDict and LazyCastedDict can be used to create dictionaries with
        automatic type casting.
    - UniqueList ensures all elements are unique and can raise an error on
        duplicates.
    - SliceableDeque extends deque with slicing support and enhanced equality
        checks.

Examples:
    >>> d = CastedDict(int, int)
    >>> d[1] = 2
    >>> d['3'] = '4'
    >>> d.update({'5': '6'})
    >>> d.update([('7', '8')])
    >>> d
    {1: 2, 3: 4, 5: 6, 7: 8}

    >>> l = UniqueList(1, 2, 3)
    >>> l.append(4)
    >>> l.append(4)
    >>> l.insert(0, 4)
    >>> l.insert(0, 5)
    >>> l[1] = 10
    >>> l
    [5, 10, 2, 3, 4]

    >>> d = SliceableDeque([1, 2, 3, 4, 5])
    >>> d[1:4]
    SliceableDeque([2, 3, 4])
"""

# pyright: reportIncompatibleMethodOverride=false
import abc
import collections
import typing

from . import types

if typing.TYPE_CHECKING:
    import _typeshed  # noqa: F401

#: A type alias for a type that can be used as a key in a dictionary.
KT = types.TypeVar('KT')
#: A type alias for a type that can be used as a value in a dictionary.
VT = types.TypeVar('VT')
#: A type alias for a dictionary with keys of type KT and values of type VT.
DT = types.Dict[KT, VT]
#: A type alias for the casted type of a dictionary key.
KT_cast = types.Optional[types.Callable[..., KT]]
#: A type alias for the casted type of a dictionary value.
VT_cast = types.Optional[types.Callable[..., VT]]
#: A type alias for the hashable values of the `UniqueList`
HT = types.TypeVar('HT', bound=types.Hashable)
#: A type alias for a regular generic type
T = types.TypeVar('T')

# Using types.Union instead of | since Python 3.7 doesn't fully support it
DictUpdateArgs = types.Union[
    types.Mapping[KT, VT],
    types.Iterable[types.Tuple[KT, VT]],
    types.Iterable[types.Mapping[KT, VT]],
    '_typeshed.SupportsKeysAndGetItem[KT, VT]',
]

OnDuplicate = types.Literal['ignore', 'raise']


class CastedDictBase(types.Dict[KT, VT], abc.ABC):
    """
    Abstract base class for dictionaries that cast keys and values.

    Attributes:
        _key_cast (KT_cast[KT]): Callable to cast dictionary keys.
        _value_cast (VT_cast[VT]): Callable to cast dictionary values.

    Methods:
        __init__(key_cast: KT_cast[KT] = None, value_cast: VT_cast[VT] = None,
            *args: DictUpdateArgs[KT, VT], **kwargs: VT) -> None:
            Initializes the dictionary with optional key and value casting
            callables.
        update(*args: DictUpdateArgs[types.Any, types.Any],
            **kwargs: types.Any) -> None:
            Updates the dictionary with the given arguments.
        __setitem__(key: types.Any, value: types.Any) -> None:
            Sets the item in the dictionary, casting the key if a key cast
            callable is provided.
    """

    _key_cast: KT_cast[KT]
    _value_cast: VT_cast[VT]

    def __init__(
        self,
        key_cast: KT_cast[KT] = None,
        value_cast: VT_cast[VT] = None,
        *args: DictUpdateArgs[KT, VT],
        **kwargs: VT,
    ) -> None:
        """
        Initializes the CastedDictBase with optional key and value
        casting callables.

        Args:
            key_cast (KT_cast[KT], optional): Callable to cast
                dictionary keys. Defaults to None.
            value_cast (VT_cast[VT], optional): Callable to cast
                dictionary values. Defaults to None.
            *args (DictUpdateArgs[KT, VT]): Arguments to initialize
                the dictionary.
            **kwargs (VT): Keyword arguments to initialize the
                dictionary.
        """
        self._value_cast = value_cast
        self._key_cast = key_cast
        self.update(*args, **kwargs)

    def update(
        self, *args: DictUpdateArgs[types.Any, types.Any], **kwargs: types.Any
    ) -> None:
        """
        Updates the dictionary with the given arguments.

        Args:
            *args (DictUpdateArgs[types.Any, types.Any]): Arguments to update
                the dictionary.
            **kwargs (types.Any): Keyword arguments to update the dictionary.
        """
        if args:
            kwargs.update(*args)

        if kwargs:
            for key, value in kwargs.items():
                self[key] = value

    def __setitem__(self, key: types.Any, value: types.Any) -> None:
        """
        Sets the item in the dictionary, casting the key if a key cast
        callable is provided.

        Args:
            key (types.Any): The key to set in the dictionary.
            value (types.Any): The value to set in the dictionary.
        """
        if self._key_cast is not None:
            key = self._key_cast(key)

        return super().__setitem__(key, value)


class CastedDict(CastedDictBase[KT, VT]):
    """
    Custom dictionary that casts keys and values to the specified typing.

    Note that you can specify the types for mypy and type hinting with:
    CastedDict[int, int](int, int)

    >>> d: CastedDict[int, int] = CastedDict(int, int)
    >>> d[1] = 2
    >>> d['3'] = '4'
    >>> d.update({'5': '6'})
    >>> d.update([('7', '8')])
    >>> d
    {1: 2, 3: 4, 5: 6, 7: 8}
    >>> list(d.keys())
    [1, 3, 5, 7]
    >>> list(d)
    [1, 3, 5, 7]
    >>> list(d.values())
    [2, 4, 6, 8]
    >>> list(d.items())
    [(1, 2), (3, 4), (5, 6), (7, 8)]
    >>> d[3]
    4

    # Casts are optional and can be disabled by passing None as the cast
    >>> d = CastedDict()
    >>> d[1] = 2
    >>> d['3'] = '4'
    >>> d.update({'5': '6'})
    >>> d.update([('7', '8')])
    >>> d
    {1: 2, '3': '4', '5': '6', '7': '8'}
    """

    def __setitem__(self, key: typing.Any, value: typing.Any) -> None:
        """Sets `key` to `cast(value)` in the dictionary."""
        if self._value_cast is not None:
            value = self._value_cast(value)

        super().__setitem__(key, value)


class LazyCastedDict(CastedDictBase[KT, VT]):
    """
    Custom dictionary that casts keys and lazily casts values to the specified
    typing. Note that the values are cast only when they are accessed and
    are not cached between executions.

    Note that you can specify the types for mypy and type hinting with:
    LazyCastedDict[int, int](int, int)

    >>> d: LazyCastedDict[int, int] = LazyCastedDict(int, int)
    >>> d[1] = 2
    >>> d['3'] = '4'
    >>> d.update({'5': '6'})
    >>> d.update([('7', '8')])
    >>> d
    {1: 2, 3: '4', 5: '6', 7: '8'}
    >>> list(d.keys())
    [1, 3, 5, 7]
    >>> list(d)
    [1, 3, 5, 7]
    >>> list(d.values())
    [2, 4, 6, 8]
    >>> list(d.items())
    [(1, 2), (3, 4), (5, 6), (7, 8)]
    >>> d[3]
    4

    # Casts are optional and can be disabled by passing None as the cast
    >>> d = LazyCastedDict()
    >>> d[1] = 2
    >>> d['3'] = '4'
    >>> d.update({'5': '6'})
    >>> d.update([('7', '8')])
    >>> d
    {1: 2, '3': '4', '5': '6', '7': '8'}
    >>> list(d.keys())
    [1, '3', '5', '7']
    >>> list(d.values())
    [2, '4', '6', '8']

    >>> list(d.items())
    [(1, 2), ('3', '4'), ('5', '6'), ('7', '8')]
    >>> d['3']
    '4'
    """

    def __setitem__(self, key: types.Any, value: types.Any) -> None:
        """
        Sets the item in the dictionary, casting the key if a key cast
        callable is provided.

        Args:
            key (types.Any): The key to set in the dictionary.
            value (types.Any): The value to set in the dictionary.
        """
        if self._key_cast is not None:
            key = self._key_cast(key)

        super().__setitem__(key, value)

    def __getitem__(self, key: types.Any) -> VT:
        """
        Gets the item from the dictionary, casting the value if a value cast
        callable is provided.

        Args:
            key (types.Any): The key to get from the dictionary.

        Returns:
            VT: The value from the dictionary.
        """
        if self._key_cast is not None:
            key = self._key_cast(key)

        value = super().__getitem__(key)

        if self._value_cast is not None:
            value = self._value_cast(value)

        return value

    def items(  # type: ignore[override]
        self,
    ) -> types.Generator[types.Tuple[KT, VT], None, None]:
        """
        Returns a generator of the dictionary's items, casting the values if a
        value cast callable is provided.

        Yields:
            types.Generator[types.Tuple[KT, VT], None, None]: A generator of
                the dictionary's items.
        """
        if self._value_cast is None:
            yield from super().items()
        else:
            for key, value in super().items():
                yield key, self._value_cast(value)

    def values(self) -> types.Generator[VT, None, None]:  # type: ignore[override]
        """
        Returns a generator of the dictionary's values, casting the values if a
        value cast callable is provided.

        Yields:
            types.Generator[VT, None, None]: A generator of the dictionary's
                values.
        """
        if self._value_cast is None:
            yield from super().values()
        else:
            for value in super().values():
                yield self._value_cast(value)


class UniqueList(types.List[HT]):
    """
    A list that only allows unique values. Duplicate values are ignored by
    default, but can be configured to raise an exception instead.

    >>> l = UniqueList(1, 2, 3)
    >>> l.append(4)
    >>> l.append(4)
    >>> l.insert(0, 4)
    >>> l.insert(0, 5)
    >>> l[1] = 10
    >>> l
    [5, 10, 2, 3, 4]

    >>> l = UniqueList(1, 2, 3, on_duplicate='raise')
    >>> l.append(4)
    >>> l.append(4)
    Traceback (most recent call last):
    ...
    ValueError: Duplicate value: 4
    >>> l.insert(0, 4)
    Traceback (most recent call last):
    ...
    ValueError: Duplicate value: 4
    >>> 4 in l
    True
    >>> l[0]
    1
    >>> l[1] = 4
    Traceback (most recent call last):
    ...
    ValueError: Duplicate value: 4
    """

    _set: types.Set[HT]

    def __init__(
        self,
        *args: HT,
        on_duplicate: OnDuplicate = 'ignore',
    ):
        """
        Initializes the UniqueList with optional duplicate handling behavior.

        Args:
            *args (HT): Initial values for the list.
            on_duplicate (OnDuplicate, optional): Behavior on duplicates.
                Defaults to 'ignore'.
        """
        self.on_duplicate = on_duplicate
        self._set = set()
        super().__init__()
        for arg in args:
            self.append(arg)

    def insert(self, index: types.SupportsIndex, value: HT) -> None:
        """
        Inserts a value at the specified index, ensuring uniqueness.

        Args:
            index (types.SupportsIndex): The index to insert the value at.
            value (HT): The value to insert.

        Raises:
            ValueError: If the value is a duplicate and `on_duplicate` is set
                to 'raise'.
        """
        if value in self._set:
            if self.on_duplicate == 'raise':
                raise ValueError(f'Duplicate value: {value}')
            else:
                return

        self._set.add(value)
        super().insert(index, value)

    def append(self, value: HT) -> None:
        """
        Appends a value to the list, ensuring uniqueness.

        Args:
            value (HT): The value to append.

        Raises:
            ValueError: If the value is a duplicate and `on_duplicate` is set
                to 'raise'.
        """
        if value in self._set:
            if self.on_duplicate == 'raise':
                raise ValueError(f'Duplicate value: {value}')
            else:
                return

        self._set.add(value)
        super().append(value)

    def __contains__(self, item: HT) -> bool:  # type: ignore[override]
        """
        Checks if the list contains the specified item.

        Args:
            item (HT): The item to check for.

        Returns:
            bool: True if the item is in the list, False otherwise.
        """
        return item in self._set

    @typing.overload
    def __setitem__(
        self, indices: types.SupportsIndex, values: HT
    ) -> None: ...

    @typing.overload
    def __setitem__(
        self, indices: slice, values: types.Iterable[HT]
    ) -> None: ...

    def __setitem__(
        self,
        indices: types.Union[slice, types.SupportsIndex],
        values: types.Union[types.Iterable[HT], HT],
    ) -> None:
        """
        Sets the item(s) at the specified index/indices, ensuring uniqueness.

        Args:
            indices (types.Union[slice, types.SupportsIndex]): The index or
                slice to set the value(s) at.
            values (types.Union[types.Iterable[HT], HT]): The value(s) to set.

        Raises:
            RuntimeError: If `on_duplicate` is 'ignore' and setting slices.
            ValueError: If the value(s) are duplicates and `on_duplicate` is
                set to 'raise'.
        """
        if isinstance(indices, slice):
            values = types.cast(types.Iterable[HT], values)
            if self.on_duplicate == 'ignore':
                raise RuntimeError(
                    'ignore mode while setting slices introduces ambiguous '
                    'behaviour and is therefore not supported'
                )

            duplicates: types.Set[HT] = set(values) & self._set
            if duplicates and values != list(self[indices]):
                raise ValueError(f'Duplicate values: {duplicates}')

            self._set.update(values)
        else:
            values = types.cast(HT, values)
            if values in self._set and values != self[indices]:
                if self.on_duplicate == 'raise':
                    raise ValueError(f'Duplicate value: {values}')
                else:
                    return

            self._set.add(values)

        super().__setitem__(
            types.cast(slice, indices), types.cast(types.List[HT], values)
        )

    def __delitem__(
        self, index: types.Union[types.SupportsIndex, slice]
    ) -> None:
        """
        Deletes the item(s) at the specified index/indices.

        Args:
            index (types.Union[types.SupportsIndex, slice]): The index or slice
                to delete the item(s) at.
        """
        if isinstance(index, slice):
            for value in self[index]:
                self._set.remove(value)
        else:
            self._set.remove(self[index])

        super().__delitem__(index)


# Type hinting `collections.deque` does not work consistently between Python
# runtime, mypy and pyright currently so we have to ignore the errors
class SliceableDeque(types.Generic[T], collections.deque[T]):
    """
    A deque that supports slicing and enhanced equality checks.

    Methods:
        __getitem__(index: types.Union[types.SupportsIndex, slice]) ->
            types.Union[T, 'SliceableDeque[T]']:
            Returns the item or slice at the given index.
        __eq__(other: types.Any) -> bool:
            Checks equality with another object, allowing for comparison with
             lists, tuples, and sets.
        pop(index: int = -1) -> T:
            Removes and returns the item at the given index. Only supports
            index 0 and the last index.
    """

    @typing.overload
    def __getitem__(self, index: types.SupportsIndex) -> T: ...

    @typing.overload
    def __getitem__(self, index: slice) -> 'SliceableDeque[T]': ...

    def __getitem__(
        self, index: types.Union[types.SupportsIndex, slice]
    ) -> types.Union[T, 'SliceableDeque[T]']:
        """
        Return the item or slice at the given index.

        Args:
            index (types.Union[types.SupportsIndex, slice]): The index or
             slice to retrieve.

        Returns:
            types.Union[T, 'SliceableDeque[T]']: The item or slice at the
            given index.

        Examples:
            >>> d = SliceableDeque[int]([1, 2, 3, 4, 5])
            >>> d[1:4]
            SliceableDeque([2, 3, 4])

            >>> d = SliceableDeque[str](['a', 'b', 'c'])
            >>> d[-2:]
            SliceableDeque(['b', 'c'])
        """
        if isinstance(index, slice):
            start, stop, step = index.indices(len(self))
            return self.__class__(self[i] for i in range(start, stop, step))
        else:
            return super().__getitem__(index)

    def __eq__(self, other: types.Any) -> bool:
        """
        Checks equality with another object, allowing for comparison with
        lists, tuples, and sets.

        Args:
            other (types.Any): The object to compare with.

        Returns:
            bool: True if the objects are equal, False otherwise.
        """
        if isinstance(other, list):
            return list(self) == other
        elif isinstance(other, tuple):
            return tuple(self) == other
        elif isinstance(other, set):
            return set(self) == other
        else:
            return super().__eq__(other)

    def pop(self, index: int = -1) -> T:
        """
        Removes and returns the item at the given index. Only supports index 0
        and the last index.

        Args:
            index (int, optional): The index of the item to remove. Defaults to
            -1.

        Returns:
            T: The removed item.

        Raises:
            IndexError: If the index is not 0 or the last index.

        Examples:
            >>> d = SliceableDeque([1, 2, 3])
            >>> d.pop(0)
            1
            >>> d.pop()
            3
        """
        if index == 0:
            return super().popleft()
        elif index in {-1, len(self) - 1}:
            return super().pop()
        else:
            raise IndexError(
                'Only index 0 and the last index (`N-1` or `-1`) '
                'are supported'
            )


if __name__ == '__main__':
    import doctest

    doctest.testmod()



"""
This module provides utilities for importing modules and handling exceptions.

Classes:
    DummyError(Exception):
        A custom exception class used as a default for exception handling.

Functions:
    import_global(name, modules=None, exceptions=DummyError, locals_=None,
        globals_=None, level=-1):
        Imports the requested items into the global scope, with support for
        relative imports and custom exception handling.
"""

from . import types


class DummyError(Exception):
    """A custom exception class used as a default for exception handling."""


# Legacy alias for DummyError
DummyException = DummyError


def import_global(  # noqa: C901
    name: str,
    modules: types.Optional[types.List[str]] = None,
    exceptions: types.ExceptionsType = DummyError,
    locals_: types.OptionalScope = None,
    globals_: types.OptionalScope = None,
    level: int = -1,
) -> types.Any:  # sourcery skip: hoist-if-from-if
    """Import the requested items into the global scope.

    WARNING! this method _will_ overwrite your global scope
    If you have a variable named `path` and you call `import_global('sys')`
    it will be overwritten with `sys.path`

    Args:
        name (str): the name of the module to import, e.g. sys
        modules (str): the modules to import, use None for everything
        exceptions (Exception): the exception to catch, e.g. ImportError
        locals_: the `locals()` method (in case you need a different scope)
        globals_: the `globals()` method (in case you need a different scope)
        level (int): the level to import from, this can be used for
        relative imports
    """
    frame = None
    name_parts: types.List[str] = name.split('.')
    modules_set: types.Set[str] = set()
    try:
        # If locals_ or globals_ are not given, autodetect them by inspecting
        # the current stack
        if locals_ is None or globals_ is None:
            import inspect

            frame = inspect.stack()[1][0]

            if locals_ is None:
                locals_ = frame.f_locals

            if globals_ is None:
                globals_ = frame.f_globals

        try:
            # Relative imports are supported (from .spam import eggs)
            if not name_parts[0]:
                name_parts = name_parts[1:]
                level = 1

            # raise IOError((name, level))
            module = __import__(
                name=name_parts[0] or '.',
                globals=globals_,
                locals=locals_,
                fromlist=name_parts[1:],
                level=max(level, 0),
            )

            # Make sure we get the right part of a dotted import (i.e.
            # spam.eggs should return eggs, not spam)
            try:
                for attr in name_parts[1:]:
                    module = getattr(module, attr)
            except AttributeError as e:
                raise ImportError(
                    'No module named ' + '.'.join(name_parts)
                ) from e

            # If no list of modules is given, autodetect from either __all__
            # or a dir() of the module
            if not modules:
                modules_set = set(getattr(module, '__all__', dir(module)))
            else:
                modules_set = set(modules).intersection(dir(module))

            # Add all items in modules to the global scope
            for k in set(dir(module)).intersection(modules_set):
                if k and k[0] != '_':
                    globals_[k] = getattr(module, k)
        except exceptions as e:
            return e
    finally:
        # Clean up, just to be sure
        del (
            name,
            name_parts,
            modules,
            modules_set,
            exceptions,
            locals_,
            globals_,
            frame,
        )



"""
This module provides various utility decorators for Python functions
and methods.

The decorators include:

1. `set_attributes`: Sets attributes on functions and classes.
2. `listify`: Converts any generator to a list or other collection.
3. `sample`: Limits calls to a function based on a sample rate.
4. `wraps_classmethod`: Wraps classmethods with type info from a
   regular method.

Each decorator is designed to enhance the functionality of Python
functions and methods in a simple and reusable manner.
"""

import contextlib
import functools
import logging
import random

from . import types

_T = types.TypeVar('_T')
_P = types.ParamSpec('_P')
_S = types.TypeVar('_S', covariant=True)


def set_attributes(**kwargs: types.Any) -> types.Callable[..., types.Any]:
    """Decorator to set attributes on functions and classes.

    A common usage for this pattern is the Django Admin where
    functions can get an optional short_description. To illustrate:

    Example from the Django admin using this decorator:
    https://docs.djangoproject.com/en/3.0/ref/contrib/admin/#django.contrib.admin.ModelAdmin.list_display

    Our simplified version:

    >>> @set_attributes(short_description='Name')
    ... def upper_case_name(self, obj):
    ...     return ('%s %s' % (obj.first_name, obj.last_name)).upper()

    The standard Django version:

    >>> def upper_case_name(obj):
    ...     return ('%s %s' % (obj.first_name, obj.last_name)).upper()

    >>> upper_case_name.short_description = 'Name'

    """

    def _set_attributes(
        function: types.Callable[_P, _T],
    ) -> types.Callable[_P, _T]:
        for key, value in kwargs.items():
            setattr(function, key, value)
        return function

    return _set_attributes


def listify(
    collection: types.Callable[
        [types.Iterable[_T]], types.Collection[_T]
    ] = list,
    allow_empty: bool = True,
) -> types.Callable[
    [types.Callable[..., types.Optional[types.Iterable[_T]]]],
    types.Callable[..., types.Collection[_T]],
]:
    """
    Convert any generator to a list or other type of collection.

    >>> @listify()
    ... def generator():
    ...     yield 1
    ...     yield 2
    ...     yield 3

    >>> generator()
    [1, 2, 3]

    >>> @listify()
    ... def empty_generator():
    ...     pass

    >>> empty_generator()
    []

    >>> @listify(allow_empty=False)
    ... def empty_generator_not_allowed():
    ...     pass

    >>> empty_generator_not_allowed()  # doctest: +ELLIPSIS
    Traceback (most recent call last):
    ...
    TypeError: ... `allow_empty` is `False`

    >>> @listify(collection=set)
    ... def set_generator():
    ...     yield 1
    ...     yield 1
    ...     yield 2

    >>> set_generator()
    {1, 2}

    >>> @listify(collection=dict)
    ... def dict_generator():
    ...     yield 'a', 1
    ...     yield 'b', 2

    >>> dict_generator()
    {'a': 1, 'b': 2}
    """

    def _listify(
        function: types.Callable[..., types.Optional[types.Iterable[_T]]],
    ) -> types.Callable[..., types.Collection[_T]]:
        def __listify(
            *args: types.Any, **kwargs: types.Any
        ) -> types.Collection[_T]:
            result: types.Optional[types.Iterable[_T]] = function(
                *args, **kwargs
            )
            if result is None:
                if allow_empty:
                    return collection(iter(()))
                else:
                    raise TypeError(
                        f'{function} returned `None` and `allow_empty` '
                        'is `False`'
                    )
            else:
                return collection(result)

        return __listify

    return _listify


def sample(
    sample_rate: float,
) -> types.Callable[
    [types.Callable[_P, _T]],
    types.Callable[_P, types.Optional[_T]],
]:
    """
    Limit calls to a function based on given sample rate.
    Number of calls to the function will be roughly equal to
    sample_rate percentage.

    Usage:

    >>> @sample(0.5)
    ... def demo_function(*args, **kwargs):
    ...     return 1

    Calls to *demo_function* will be limited to 50% approximatly.
    """

    def _sample(
        function: types.Callable[_P, _T],
    ) -> types.Callable[_P, types.Optional[_T]]:
        @functools.wraps(function)
        def __sample(
            *args: _P.args, **kwargs: _P.kwargs
        ) -> types.Optional[_T]:
            if random.random() < sample_rate:
                return function(*args, **kwargs)
            else:
                logging.debug(
                    'Skipped execution of %r(%r, %r) due to sampling',
                    function,
                    args,
                    kwargs,
                )
                return None

        return __sample

    return _sample


def wraps_classmethod(
    wrapped: types.Callable[types.Concatenate[_S, _P], _T],
) -> types.Callable[
    [
        types.Callable[types.Concatenate[types.Any, _P], _T],
    ],
    types.Callable[types.Concatenate[_S, _P], _T],
]:
    """
    Like `functools.wraps`, but for wrapping classmethods with the type info
    from a regular method.
    """

    def _wraps_classmethod(
        wrapper: types.Callable[types.Concatenate[types.Any, _P], _T],
    ) -> types.Callable[types.Concatenate[_S, _P], _T]:
        # For some reason `functools.update_wrapper` fails on some test
        # runs but not while running actual code
        with contextlib.suppress(AttributeError):
            wrapper = functools.update_wrapper(
                wrapper,
                wrapped,
                assigned=tuple(
                    a
                    for a in functools.WRAPPER_ASSIGNMENTS
                    if a != '__annotations__'
                ),
            )
        if annotations := getattr(wrapped, '__annotations__', {}):
            annotations.pop('self', None)
            wrapper.__annotations__ = annotations

        return wrapper

    return _wraps_classmethod



"""
This module provides utility functions for handling time-related operations.

Functions:
- timedelta_to_seconds: Convert a timedelta to seconds with microseconds as
  fraction.
- delta_to_seconds: Convert a timedelta or numeric interval to seconds.
- delta_to_seconds_or_none: Convert a timedelta to seconds or return None.
- format_time: Format a timestamp (timedelta, datetime, or seconds) to a
  string.
- timeout_generator: Generate items from an iterable until a timeout is
  reached.
- aio_timeout_generator: Asynchronously generate items from an iterable until a
  timeout is reached.
- aio_generator_timeout_detector: Detect if an async generator has not yielded
  an element for a set amount of time.
- aio_generator_timeout_detector_decorator: Decorator for
  aio_generator_timeout_detector.
"""

# pyright: reportUnnecessaryIsInstance=false
import asyncio
import datetime
import functools
import itertools
import time

import python_utils
from python_utils import aio, exceptions, types

_T = types.TypeVar('_T')
_P = types.ParamSpec('_P')


# There might be a better way to get the epoch with tzinfo, please create
# a pull request if you know a better way that functions for Python 2 and 3
epoch = datetime.datetime(year=1970, month=1, day=1)


def timedelta_to_seconds(delta: datetime.timedelta) -> types.Number:
    """Convert a timedelta to seconds with the microseconds as fraction.

    Note that this method has become largely obsolete with the
    `timedelta.total_seconds()` method introduced in Python 2.7.

    >>> from datetime import timedelta
    >>> '%d' % timedelta_to_seconds(timedelta(days=1))
    '86400'
    >>> '%d' % timedelta_to_seconds(timedelta(seconds=1))
    '1'
    >>> '%.6f' % timedelta_to_seconds(timedelta(seconds=1, microseconds=1))
    '1.000001'
    >>> '%.6f' % timedelta_to_seconds(timedelta(microseconds=1))
    '0.000001'
    """
    # Only convert to float if needed
    if delta.microseconds:
        total = delta.microseconds * 1e-6
    else:
        total = 0
    total += delta.seconds
    total += delta.days * 60 * 60 * 24
    return total


def delta_to_seconds(interval: types.delta_type) -> types.Number:
    """
    Convert a timedelta to seconds.

    >>> delta_to_seconds(datetime.timedelta(seconds=1))
    1
    >>> delta_to_seconds(datetime.timedelta(seconds=1, microseconds=1))
    1.000001
    >>> delta_to_seconds(1)
    1
    >>> delta_to_seconds('whatever')  # doctest: +ELLIPSIS
    Traceback (most recent call last):
        ...
    TypeError: Unknown type ...
    """
    if isinstance(interval, datetime.timedelta):
        return timedelta_to_seconds(interval)
    elif isinstance(interval, (int, float)):
        return interval
    else:
        raise TypeError(f'Unknown type {type(interval)}: {interval!r}')


def delta_to_seconds_or_none(
    interval: types.Optional[types.delta_type],
) -> types.Optional[types.Number]:
    """Convert a timedelta to seconds or return None."""
    if interval is None:
        return None
    else:
        return delta_to_seconds(interval)


def format_time(
    timestamp: types.timestamp_type,
    precision: datetime.timedelta = datetime.timedelta(seconds=1),
) -> str:
    """Formats timedelta/datetime/seconds.

    >>> format_time('1')
    '0:00:01'
    >>> format_time(1.234)
    '0:00:01'
    >>> format_time(1)
    '0:00:01'
    >>> format_time(datetime.datetime(2000, 1, 2, 3, 4, 5, 6))
    '2000-01-02 03:04:05'
    >>> format_time(datetime.date(2000, 1, 2))
    '2000-01-02'
    >>> format_time(datetime.timedelta(seconds=3661))
    '1:01:01'
    >>> format_time(None)
    '--:--:--'
    >>> format_time(format_time)  # doctest: +ELLIPSIS
    Traceback (most recent call last):
        ...
    TypeError: Unknown type ...

    """
    precision_seconds = precision.total_seconds()

    if isinstance(timestamp, str):
        timestamp = float(timestamp)

    if isinstance(timestamp, (int, float)):
        try:
            timestamp = datetime.timedelta(seconds=timestamp)
        except OverflowError:  # pragma: no cover
            timestamp = None

    if isinstance(timestamp, datetime.timedelta):
        seconds = timestamp.total_seconds()
        # Truncate the number to the given precision
        seconds = seconds - (seconds % precision_seconds)

        return str(datetime.timedelta(seconds=seconds))
    elif isinstance(timestamp, datetime.datetime):  # pragma: no cover
        # Python 2 doesn't have the timestamp method
        if hasattr(timestamp, 'timestamp'):
            seconds = timestamp.timestamp()
        else:
            seconds = timedelta_to_seconds(timestamp - epoch)

        # Truncate the number to the given precision
        seconds = seconds - (seconds % precision_seconds)

        try:  # pragma: no cover
            dt = datetime.datetime.fromtimestamp(seconds)
        except (ValueError, OSError):  # pragma: no cover
            dt = datetime.datetime.max
        return str(dt)
    elif isinstance(timestamp, datetime.date):
        return str(timestamp)
    elif timestamp is None:
        return '--:--:--'
    else:
        raise TypeError(f'Unknown type {type(timestamp)}: {timestamp!r}')


@types.overload
def _to_iterable(
    iterable: types.Union[
        types.Callable[[], types.AsyncIterable[_T]],
        types.AsyncIterable[_T],
    ],
) -> types.AsyncIterable[_T]: ...


@types.overload
def _to_iterable(
    iterable: types.Union[
        types.Callable[[], types.Iterable[_T]], types.Iterable[_T]
    ],
) -> types.Iterable[_T]: ...


def _to_iterable(
    iterable: types.Union[
        types.Iterable[_T],
        types.Callable[[], types.Iterable[_T]],
        types.AsyncIterable[_T],
        types.Callable[[], types.AsyncIterable[_T]],
    ],
) -> types.Union[types.Iterable[_T], types.AsyncIterable[_T]]:
    if callable(iterable):
        return iterable()
    else:
        return iterable


def timeout_generator(
    timeout: types.delta_type,
    interval: types.delta_type = datetime.timedelta(seconds=1),
    iterable: types.Union[
        types.Iterable[_T], types.Callable[[], types.Iterable[_T]]
    ] = itertools.count,  # type: ignore[assignment]
    interval_multiplier: float = 1.0,
    maximum_interval: types.Optional[types.delta_type] = None,
) -> types.Iterable[_T]:
    """
    Generator that walks through the given iterable (a counter by default)
    until the float_timeout is reached with a configurable float_interval
    between items.

    This can be used to limit the time spent on a slow operation. This can be
    useful for testing slow APIs so you get a small sample of the data in a
    reasonable amount of time.

    >>> for i in timeout_generator(0.1, 0.06):
    ...     # Put your slow code here
    ...     print(i)
    0
    1
    2
    >>> timeout = datetime.timedelta(seconds=0.1)
    >>> interval = datetime.timedelta(seconds=0.06)
    >>> for i in timeout_generator(timeout, interval, itertools.count()):
    ...     print(i)
    0
    1
    2
    >>> for i in timeout_generator(1, interval=0.1, iterable='ab'):
    ...     print(i)
    a
    b

    >>> timeout = datetime.timedelta(seconds=0.1)
    >>> interval = datetime.timedelta(seconds=0.06)
    >>> for i in timeout_generator(timeout, interval, interval_multiplier=2):
    ...     print(i)
    0
    1
    2
    """
    float_interval: float = delta_to_seconds(interval)
    float_maximum_interval: types.Optional[float] = delta_to_seconds_or_none(
        maximum_interval
    )
    iterable_ = _to_iterable(iterable)

    end = delta_to_seconds(timeout) + time.perf_counter()
    for item in iterable_:
        yield item

        if time.perf_counter() >= end:
            break

        time.sleep(float_interval)

        float_interval *= interval_multiplier
        if float_maximum_interval:
            float_interval = min(float_interval, float_maximum_interval)


async def aio_timeout_generator(
    timeout: types.delta_type,  # noqa: ASYNC109
    interval: types.delta_type = datetime.timedelta(seconds=1),
    iterable: types.Union[
        types.AsyncIterable[_T], types.Callable[..., types.AsyncIterable[_T]]
    ] = aio.acount,
    interval_multiplier: float = 1.0,
    maximum_interval: types.Optional[types.delta_type] = None,
) -> types.AsyncGenerator[_T, None]:
    """
    Async generator that walks through the given async iterable (a counter by
    default) until the float_timeout is reached with a configurable
    float_interval between items.

    The interval_exponent automatically increases the float_timeout with each
    run. Note that if the float_interval is less than 1, 1/interval_exponent
    will be used so the float_interval is always growing. To double the
    float_interval with each run, specify 2.

    Doctests and asyncio are not friends, so no examples. But this function is
    effectively the same as the `timeout_generator` but it uses `async for`
    instead.
    """
    float_interval: float = delta_to_seconds(interval)
    float_maximum_interval: types.Optional[float] = delta_to_seconds_or_none(
        maximum_interval
    )
    iterable_ = _to_iterable(iterable)

    end = delta_to_seconds(timeout) + time.perf_counter()
    async for item in iterable_:  # pragma: no branch
        yield item

        if time.perf_counter() >= end:
            break

        await asyncio.sleep(float_interval)

        float_interval *= interval_multiplier
        if float_maximum_interval:  # pragma: no branch
            float_interval = min(float_interval, float_maximum_interval)


async def aio_generator_timeout_detector(
    generator: types.AsyncGenerator[_T, None],
    timeout: types.Optional[types.delta_type] = None,  # noqa: ASYNC109
    total_timeout: types.Optional[types.delta_type] = None,
    on_timeout: types.Optional[
        types.Callable[
            [
                types.AsyncGenerator[_T, None],
                types.Optional[types.delta_type],
                types.Optional[types.delta_type],
                BaseException,
            ],
            types.Any,
        ]
    ] = exceptions.reraise,
    **on_timeout_kwargs: types.Mapping[types.Text, types.Any],
) -> types.AsyncGenerator[_T, None]:
    """
    This function is used to detect if an asyncio generator has not yielded
    an element for a set amount of time.

    The `on_timeout` argument is called with the `generator`, `timeout`,
    `total_timeout`, `exception` and the extra `**kwargs` to this function as
    arguments.
    If `on_timeout` is not specified, the exception is reraised.
    If `on_timeout` is `None`, the exception is silently ignored and the
    generator will finish as normal.
    """
    if total_timeout is None:
        total_timeout_end = None
    else:
        total_timeout_end = time.perf_counter() + delta_to_seconds(
            total_timeout
        )

    timeout_s = python_utils.delta_to_seconds_or_none(timeout)

    while True:
        try:
            if total_timeout_end and time.perf_counter() >= total_timeout_end:
                raise asyncio.TimeoutError(  # noqa: TRY301
                    'Total timeout reached'
                )

            if timeout_s:
                yield await asyncio.wait_for(generator.__anext__(), timeout_s)
            else:
                yield await generator.__anext__()

        except asyncio.TimeoutError as exception:  # noqa: PERF203
            if on_timeout is not None:
                await on_timeout(
                    generator,
                    timeout,
                    total_timeout,
                    exception,
                    **on_timeout_kwargs,
                )
            break

        except StopAsyncIteration:
            break


def aio_generator_timeout_detector_decorator(
    timeout: types.Optional[types.delta_type] = None,
    total_timeout: types.Optional[types.delta_type] = None,
    on_timeout: types.Optional[
        types.Callable[
            [
                types.AsyncGenerator[types.Any, None],
                types.Optional[types.delta_type],
                types.Optional[types.delta_type],
                BaseException,
            ],
            types.Any,
        ]
    ] = exceptions.reraise,
    **on_timeout_kwargs: types.Mapping[types.Text, types.Any],
) -> types.Callable[
    [types.Callable[_P, types.AsyncGenerator[_T, None]]],
    types.Callable[_P, types.AsyncGenerator[_T, None]],
]:
    """A decorator wrapper for aio_generator_timeout_detector."""

    def _timeout_detector_decorator(
        generator: types.Callable[_P, types.AsyncGenerator[_T, None]],
    ) -> types.Callable[_P, types.AsyncGenerator[_T, None]]:
        """The decorator itself."""

        @functools.wraps(generator)
        def wrapper(
            *args: _P.args,
            **kwargs: _P.kwargs,
        ) -> types.AsyncGenerator[_T, None]:
            return aio_generator_timeout_detector(
                generator(*args, **kwargs),
                timeout,
                total_timeout,
                on_timeout,
                **on_timeout_kwargs,
            )

        return wrapper

    return _timeout_detector_decorator



"""Asyncio equivalents to regular Python functions."""

import asyncio
import itertools
import typing

from . import types

_N = types.TypeVar('_N', int, float)
_T = types.TypeVar('_T')
_K = types.TypeVar('_K')
_V = types.TypeVar('_V')


async def acount(
    start: _N = 0,
    step: _N = 1,
    delay: float = 0,
    stop: types.Optional[_N] = None,
) -> types.AsyncIterator[_N]:
    """Asyncio version of itertools.count()."""
    for item in itertools.count(start, step):  # pragma: no branch
        if stop is not None and item >= stop:
            break

        yield item
        await asyncio.sleep(delay)


@typing.overload
async def acontainer(
    iterable: types.Union[
        types.AsyncIterable[_T],
        types.Callable[..., types.AsyncIterable[_T]],
    ],
    container: types.Type[types.Tuple[_T, ...]],
) -> types.Tuple[_T, ...]: ...


@typing.overload
async def acontainer(
    iterable: types.Union[
        types.AsyncIterable[_T],
        types.Callable[..., types.AsyncIterable[_T]],
    ],
    container: types.Type[types.List[_T]] = list,
) -> types.List[_T]: ...


@typing.overload
async def acontainer(
    iterable: types.Union[
        types.AsyncIterable[_T],
        types.Callable[..., types.AsyncIterable[_T]],
    ],
    container: types.Type[types.Set[_T]],
) -> types.Set[_T]: ...


async def acontainer(
    iterable: types.Union[
        types.AsyncIterable[_T],
        types.Callable[..., types.AsyncIterable[_T]],
    ],
    container: types.Callable[
        [types.Iterable[_T]], types.Collection[_T]
    ] = list,
) -> types.Collection[_T]:
    """
    Asyncio version of list()/set()/tuple()/etc() using an async for loop.

    So instead of doing `[item async for item in iterable]` you can do
    `await acontainer(iterable)`.

    """
    iterable_: types.AsyncIterable[_T]
    if callable(iterable):
        iterable_ = iterable()
    else:
        iterable_ = iterable

    item: _T
    items: types.List[_T] = []
    async for item in iterable_:  # pragma: no branch
        items.append(item)

    return container(items)


async def adict(
    iterable: types.Union[
        types.AsyncIterable[types.Tuple[_K, _V]],
        types.Callable[..., types.AsyncIterable[types.Tuple[_K, _V]]],
    ],
    container: types.Callable[
        [types.Iterable[types.Tuple[_K, _V]]], types.Mapping[_K, _V]
    ] = dict,
) -> types.Mapping[_K, _V]:
    """
    Asyncio version of dict() using an async for loop.

    So instead of doing `{key: value async for key, value in iterable}` you
    can do `await adict(iterable)`.

    """
    iterable_: types.AsyncIterable[types.Tuple[_K, _V]]
    if callable(iterable):
        iterable_ = iterable()
    else:
        iterable_ = iterable

    item: types.Tuple[_K, _V]
    items: types.List[types.Tuple[_K, _V]] = []
    async for item in iterable_:  # pragma: no branch
        items.append(item)

    return container(items)



"""
This module contains metadata about the `python-utils` package.

Attributes:
    __package_name__ (str): The name of the package.
    __author__ (str): The author of the package.
    __author_email__ (str): The email of the author.
    __description__ (str): A brief description of the package.
    __url__ (str): The URL of the package's repository.
    __version__ (str): The current version of the package.
"""

__package_name__: str = 'python-utils'
__author__: str = 'Rick van Hattem'
__author_email__: str = 'Wolph@wol.ph'
__description__: str = (
    'Python Utils is a module with some convenient utilities not included '
    'with the standard Python install'
)
__url__: str = 'https://github.com/WoLpH/python-utils'
# Omit type info due to automatic versioning script
__version__ = '3.9.1'



"""
This module provides a `Logurud` class that integrates the `loguru` logger
with the base logging functionality defined in `logger_module.LoggerBase`.

Classes:
    Logurud: A class that extends `LoggerBase` and uses `loguru` for logging.

Usage example:
    >>> from python_utils.loguru import Logurud
    >>> class MyClass(Logurud):
    ...     def __init__(self):
    ...         Logurud.__init__(self)
    >>> my_class = MyClass()
    >>> my_class.logger.info('This is an info message')
"""

from __future__ import annotations

import typing

import loguru

from . import logger as logger_module

__all__ = ['Logurud']


class Logurud(logger_module.LoggerBase):
    """
    A class that extends `LoggerBase` and uses `loguru` for logging.

    Attributes:
        logger (loguru.Logger): The `loguru` logger instance.
    """

    logger: loguru.Logger

    def __new__(cls, *args: typing.Any, **kwargs: typing.Any) -> Logurud:
        """
        Creates a new instance of `Logurud` and initializes the `loguru`
        logger.

        Args:
            *args (typing.Any): Variable length argument list.
            **kwargs (typing.Any): Arbitrary keyword arguments.

        Returns:
            Logurud: A new instance of `Logurud`.
        """
        cls.logger: loguru.Logger = loguru.logger.opt(depth=1)
        return super().__new__(cls)



"""
This module provides utility functions for raising and reraising exceptions.

Functions:
    raise_exception(exception_class, *args, **kwargs):
        Returns a function that raises an exception of the given type with
        the given arguments.

    reraise(*args, **kwargs):
        Reraises the current exception.
"""

from . import types


def raise_exception(
    exception_class: types.Type[Exception],
    *args: types.Any,
    **kwargs: types.Any,
) -> types.Callable[..., None]:
    """
    Returns a function that raises an exception of the given type with the
    given arguments.

    >>> raise_exception(ValueError, 'spam')('eggs')
    Traceback (most recent call last):
        ...
    ValueError: spam
    """

    def raise_(*args_: types.Any, **kwargs_: types.Any) -> types.Any:
        raise exception_class(*args, **kwargs)

    return raise_


def reraise(*args: types.Any, **kwargs: types.Any) -> types.Any:
    """
    Reraises the current exception.

    This function seems useless, but it can be useful when you need to pass
    a callable to another function that raises an exception.
    """
    raise
